
#include "globdef.h"
#include "uidef.h"
#include "sigdef.h"
#include "screendef.h"
#include "fft3def.h"
#include "options.h"
#include "thrdef.h"


void make_coherent_graph(int clear_old);
void manage_meter_graph(void);
int coh_tophat_filter(float *xin,float *xout,int lk,int fpt);
int coherent_graph_scro;
int cg_old_y1;
int cg_old_y2;
int cg_old_x1;
int cg_old_x2;

extern int meter_graph_scro;
int zzpos;

// float[2] baseb_out=data to send to loudspeaker
// float[2] baseb=complex amplitude of first level coherent data.
// float[2] baseb_raw=complex amplitude of baseband signal. Raw data, pol. adapted.
// float[2] baseb_raw_orthog=complex amplitude of pol. orthog signal. Raw data.
// float[2] baseb_carrier=phase of carrier. Complex data, cosines and sines.
// float[1] baseb_carrier_ampl=amplitude of carrier
// float[1] baseb_totpwr=total power of baseb_raw
// float[2] baseb_envelope=complex amplitude from fitted dots and dashes.
// float[1] baseb_upthreshold=forward peak detector for power
// float[1] baseb_threshold=combined forward and backward peak detector at -6dB
// float[2] baseb_fit=fitted dots and dashes.
// float[2] baseb_tmp=array for intermediate data in complex format
// float[1] baseb_agc_level=used only when AGC is enabled.
// short int[1] baseb_ramp=indicator for time of power above/below threshold.
// short_int[1] baseb_clock=CW code clock
// float[2] baseb_tmp=for debug during development

// baseb_pa  The starting point of the latest mix2_size/2 points.
// baseb_pb  The point up to which thresholds exist.
// baseb_pc  The point up to which cw speed statistics and ramp is collected.
// baseb_pd  A key up region close before baseb_pc
// baseb_pe  The point up to which we have run first detect.
// baseb_pf
// baseb_px  The oldest point that contains valid data.

#define COH_GRAPH_ACTIVE -1000
#define X_SIZE 5


void coherent(void)
{
int i,j,k;
int ix;
int last_point;
int ia,ib,lmin,lmax;
float t1;
// Make a first decision key up/key down based on signal power.
// This is the conventional approach, a matched filter (the operator
// has to select the optimum bandwidth) followed by a squarelaw detector.
// The threshold is conventionally kept at a fixed level but here it is
// very close to the noise when no signal is present. while it is at
// half the peak amplitude for strong signals.
// Store the result in baseb_ramp as ramps.
// Positive values indicate key down while negative values
// indicate key up (or signal too weak)
ib=(baseb_pb+baseband_mask)&baseband_mask;
ia=baseb_pb;
last_point=(baseb_pa-(int)(7*cg_code_unit)+baseband_size)&baseband_mask;
baseb_pb=last_point;
while( ia != last_point )
  {
  if(baseb_totpwr[ia] > baseb_threshold[ia])
    {
    if(baseb_ramp[ib]>0)
      {
      baseb_ramp[ia]=baseb_ramp[ib]+1;
      }
    else
      {
      baseb_ramp[ia]=1;
      }
    }
  else
    {
    if(baseb_ramp[ib]<0)
      {
      baseb_ramp[ia]=baseb_ramp[ib]-1;
      }
    else
      {
      k=cg_size*baseb_ramp[ib]/(6*cg_code_unit);
      if(k >= cg_size)k=cg_size-1;
      cg_statistics[k]++;
      cg_statistics_sum++;
      baseb_ramp[ia]=-1;
      }
    }
  ib=ia;
  ia=(ia+1)&baseband_mask;
  }
if(cg_statistics_sum>CG_STATISTICS_F2*cg_size)
  {
  t1=CG_STATISTICS_F2*cg_size/cg_statistics_sum;
  cg_statistics_sum=CG_STATISTICS_F2*cg_size;
  }
else
  {
  t1=1;
  }
for(ix=0; ix<cg_size; ix++)
  {
  cg_statistics[ix]*=t1;
  }
if(genparm[CW_DECODE_ENABLE] == 0)goto no_decode;
if(cw_detect_flag >= 10)
  {
  switch (cw_detect_flag)
    {
    case 10:
    init_cw_decode_region();
    break;

    case 11:
    cw_decode_region();
    break;

    case 20:
    init_cw_decode();
    break;

    case 21:
    cw_decode();
    break;
    }

  return;
  }
lmax=lmin=j=0;
// *****************************************************************
// Assuming the operator selected a near optimum filter bandwidth
// the length of a Morse code bit should be about cg_code_unit points.
// The phase can not change over such a short time because then
// the frequency would not be inside the passband.
// The Morse code long parts are about 3 times longer and in
// that time the phase may change considerably.
// We have arranged for the average phase to be zero in baseb by
// the I/Q demodulator using the carrier phase, but if the transmitter 
// is "chirpy" the frequency drifts up and down synchronous with keying.
// Accumulate the waveform each time the positive ramp is between
// 2 and 5 times cg_code_units.
// Collect the data over CWSPEED_UNITS Morse code units, then switch buffer.
last_point=(baseb_pa-(int)(10*cg_code_unit)+baseband_size)&baseband_mask;
ia=(last_point+baseband_mask)&baseband_mask;
// Step ia backwards until a long region is found. A region of more than
// 5 units is likely to be key up, but if it is key down, back across it.
j=0;
while(j<5*cg_code_unit)
  {
  i=abs(baseb_ramp[ia]);
  j+=i;
  ia=(ia-i+baseband_size)&baseband_mask;
  if( ((ia-baseb_pd+baseband_size)&baseband_mask)>baseband_neg)return;
  }
if(baseb_ramp[ia]>0)
  {
  ia=(ia-baseb_ramp[ia]+baseband_size)&baseband_mask;
  if( ((ia-baseb_pd+baseband_size)&baseband_mask)>baseband_neg)return;
  }
k=(ia-baseb_pd+baseband_size)&baseband_mask;
if(k<32*cg_code_unit)return;
// We actually have data to investigate.
cg_collect_points+=k;
lmin=2*cg_code_unit;
lmax=6*cg_code_unit;
ib=baseb_pc;
baseb_pc=ia;
while( ((ia-ib+baseband_size)&baseband_mask) <= k)
  {
  if(baseb_ramp[ia]<0)
    {
    ia=(ia+baseb_ramp[ia]+baseband_size)&baseband_mask;
    }
  else
    {
    if(baseb_ramp[ia]>lmin  &&  baseb_ramp[ia]<lmax)
      {
      add_cwlong( (ia-baseb_ramp[ia]/2)&baseband_mask);
      }
    ia=(ia-(int)baseb_ramp[ia]+baseband_size)&baseband_mask;
    }
  }
// If we averaged over the specified time, shift buffer.
if(cg_collect_points>CWSPEED_UNITS*cg_code_unit)
  {
  if(dash_waveform_established == 0)
    {
// We do not have a waveform from previous sucessful detection.
// At least one of the cwspeed buffers is filled with data.
// Extract the average shape of dot's and dashes or make
// a guess based on the selected bandwidth and statistics in 
// case data is poor.
    get_average_shapes();
    }
  cg_collect_points=0;
  cg_shape_bufno++;
  if(cg_shape_bufno >= CWLONG_MAXBUF)
    {
    cg_shape_bufno=0;
    }
  k=2*cg_shape_bufno*cwlong_store_points;
  for(i=0; i<2*cwlong_store_points; i++)cwlong_buffer[i+k]=0;
  cwlong_no_of_traces[cg_shape_bufno]=0;
  }   
// If we have collected data corresponding to several Morse code
// characters, twice the data for one cwspeed buffer, try
// to detect key down and key up 
i=2*CWSPEED_UNITS;
if(cw_detect_flag != 0)i/=8;
if( ((baseb_pb-baseb_pe+baseband_size)&baseband_mask) > i*cg_code_unit)
  {
// Place baseb_pc at the end of a key up, but make sure
// there are several valid points after baseb_pc.
  baseb_pc=(baseb_pb+baseband_mask)&baseband_mask;
  i=0;
  while(baseb_ramp[baseb_pc]>0 || i < 4*cwbit_pts)
    {
    k=abs(baseb_ramp[baseb_pc]);
    baseb_pc=(baseb_pc-k+baseband_mask)&baseband_mask;
    i+=k;
    }
  if( ((baseb_pc-baseb_pe+baseband_size)&baseband_mask) >
                                  0.2*CWSPEED_UNITS*cg_code_unit)
    {
    switch (cw_detect_flag)
      {
      case 0:
      detect_cw_speed();
      break;
        
      case 1:
      first_find_parts();
      break;

      case 2:
      second_find_parts();
      break;
      }
    }
  }
no_decode:;
}

void check_cg_borders(void)
{
int xsiz, ysiz;
xsiz=3*cg_size+2;
ysiz=xsiz+6*text_height+2;
current_graph_minh=ysiz;
current_graph_minw=xsiz;
check_graph_placement((void*)(&cg));
set_graph_minwidth((void*)(&cg));
}


void help_on_coherent_graph(void)
{
int msg_no;
int event_no;
msg_no=68;
for(event_no=0; event_no<MAX_CGBUTT; event_no++)
  {
  if( cgbutt[event_no].x1 <= mouse_x &&
      cgbutt[event_no].x2 >= mouse_x &&
      cgbutt[event_no].y1 <= mouse_y &&
      cgbutt[event_no].y2 >= mouse_y)
    {
    switch (event_no)
      {
      case CG_TOP:
      case CG_BOTTOM:
      case CG_LEFT:
      case CG_RIGHT:
      msg_no=101;
      break;

      case CG_OSCILLOSCOPE: 
      msg_no=69;
      break;

      case CG_METER_GRAPH:
      msg_no=99;
      break;      
      }
    }
  }
help_message(msg_no);
}




void mouse_continue_coh_graph(void)
{
int i, j;
switch (mouse_active_flag-1)
  {
  case CG_TOP:
  if(cg.ytop!=mouse_y)goto cgm;
  break;

  case CG_BOTTOM:
  if(cg.ybottom!=mouse_y)goto cgm;
  break;

  case CG_LEFT:
  if(cg.xleft!=mouse_x)goto cgm;
  break;

  case CG_RIGHT:
  if(cg.xright==mouse_x)break;
cgm:;
  pause_screen_and_hide_mouse();
  graph_borders((void*)&cg,0);
  if(cg_oldx==COH_GRAPH_ACTIVE)
    {
    cg_oldx=mouse_x;
    cg_oldy=mouse_y;
    }
  else
    {
    i=mouse_x-cg_oldx;
    j=mouse_y-cg_oldy;
    cg_oldx=mouse_x;
    cg_oldy=mouse_y;
    cg.ytop+=j;
    cg.ybottom+=j;
    cg.xleft+=i;
    cg.xright+=i;
    check_cg_borders();
    }
  graph_borders((void*)&cg,15);
  resume_thread(THREAD_SCREEN);
  break;
    
  default:
  goto await_release;
  }
if(leftpressed == BUTTON_RELEASED)goto finish;
return;
await_release:;
if(leftpressed != BUTTON_RELEASED)return;
switch (mouse_active_flag-1)
  {
  case CG_OSCILLOSCOPE:
  pause_thread(THREAD_SCREEN);
  hide_mouse(0,screen_width>>1,0,screen_height);
  clear_cg_traces();
  resume_thread(THREAD_SCREEN);
  cg_oscill_on^=1;
  cg_oscill_on&=1;
  cg_max_trlevel=0;
  break;

  case CG_METER_GRAPH:
  pause_thread(THREAD_SCREEN);
  hide_mouse(0,screen_width>>1,0,screen_height);
  cg.meter_graph_on^=1;
  manage_meter_graph();
  resume_thread(THREAD_SCREEN);
  break;
  
  default:
  lirerr(872);
  break;
  }
finish:;
leftpressed=BUTTON_IDLE;
mouse_active_flag=0;
make_coherent_graph(TRUE);
cg_oldx=COH_GRAPH_ACTIVE;
}


void mouse_on_coh_graph(void)
{
int event_no;
// First find out is we are on a button or border line.
for(event_no=0; event_no<MAX_CGBUTT; event_no++)
  {
  if( cgbutt[event_no].x1 <= mouse_x &&
      cgbutt[event_no].x2 >= mouse_x &&
      cgbutt[event_no].y1 <= mouse_y &&
      cgbutt[event_no].y2 >= mouse_y)
    {
    cg_old_y1=cg.ytop;
    cg_old_y2=cg.ybottom;
    cg_old_x1=cg.xleft;
    cg_old_x2=cg.xright;
    mouse_active_flag=1+event_no;
    current_mouse_activity=mouse_continue_coh_graph;
    return;
    }
  }
// Not button or border.
// Do nothing.
current_mouse_activity=mouse_nothing;
mouse_active_flag=1;
}

void make_coherent_graph(int clear_old)
{
int len;
pause_thread(THREAD_SCREEN);
hide_mouse(0,screen_width>>1,0,screen_height);
clear_cg_traces();
if(clear_old)
  {
  hide_mouse(cg_old_x1,cg_old_x2,cg_old_y1,cg_old_y2);
  lir_fillbox(cg_old_x1,cg_old_y1,cg_old_x2-cg_old_x1+1,
                                                    cg_old_y2-cg_old_y1+1,0);
  }
check_cg_borders();
hide_mouse(cg.xleft,cg.xright,cg.ytop,cg.ybottom);  
scro[coherent_graph_scro].no=COH_GRAPH;
scro[coherent_graph_scro].x1=cg.xleft;
scro[coherent_graph_scro].x2=cg.xright;
scro[coherent_graph_scro].y1=cg.ytop;
scro[coherent_graph_scro].y2=cg.ybottom;
cgbutt[CG_LEFT].x1=cg.xleft;
cgbutt[CG_LEFT].x2=cg.xleft;
cgbutt[CG_LEFT].y1=cg.ytop;
cgbutt[CG_LEFT].y2=cg.ybottom;
cgbutt[CG_RIGHT].x1=cg.xright;
cgbutt[CG_RIGHT].x2=cg.xright;
cgbutt[CG_RIGHT].y1=cg.ytop;
cgbutt[CG_RIGHT].y2=cg.ybottom;
cgbutt[CG_TOP].x1=cg.xleft;
cgbutt[CG_TOP].x2=cg.xright;
cgbutt[CG_TOP].y1=cg.ytop;
cgbutt[CG_TOP].y2=cg.ytop;
cgbutt[CG_BOTTOM].x1=cg.xleft;
cgbutt[CG_BOTTOM].x2=cg.xright;
cgbutt[CG_BOTTOM].y1=cg.ybottom;
cgbutt[CG_BOTTOM].y2=cg.ybottom;
// Draw the border lines
graph_borders((void*)&cg,7);
make_button(cg.xleft+text_width-1,cg.ybottom-text_height/2-1,
                                               cgbutt,CG_OSCILLOSCOPE,'o');
make_button(cg.xleft+text_width-1,cg.ybottom-5*(text_height/2-1),
                                               cgbutt,CG_METER_GRAPH,'s');
cg_oldx=COH_GRAPH_ACTIVE;
settextcolor(7);
cg_x0=(cg.xleft+cg.xright)>>1;
len=((cg.xright-cg.xleft)>>1);
cg_y0=cg.ytop+len+1;
cg_y1=cg_y0+len;
lir_hline(cg.xleft+1,cg_y1,cg.xright-1,7);
if(kill_all_flag) return;
clear_coherent();
manage_meter_graph();
resume_thread(THREAD_SCREEN);
make_modepar_file(GRAPHTYPE_CG);
}

void init_coherent_graph(void)
{
int xsiz, ysiz;
zzpos=2*CG_AMAX+2;
xsiz=3*cg_size+2;
ysiz=xsiz+6*text_height+2;
if (read_modepar_file(GRAPHTYPE_CG) == 0)
  {
reinit:;
  cg.xright=screen_width-1;
  cg.xleft=cg.xright-xsiz;
  cg.ybottom=bg.ybottom;
  cg.ytop=cg.ybottom-ysiz;
  cg.meter_graph_on=1;
  }
if(cg.ybottom-cg.ytop != ysiz ||
   cg.xright-cg.xleft != xsiz)goto reinit;
cg_flag=1;
mg_flag=0;
cg_oscill_on=0;
if(genparm[AFC_ENABLE] !=0 && genparm[CW_DECODE_ENABLE] != 0)cg_oscill_on=1;
cg_osc_shift_flag=0;
cg_max_trlevel=0;
coherent_graph_scro=no_of_scro;
meter_graph_scro=no_of_scro+1;
if(no_of_scro+2 >= MAX_SCRO)lirerr(89);
make_coherent_graph(FALSE);
no_of_scro+=2;
}

int coh_tophat_filter(float *xin,float *xout,int lk,int lapt)
{
int i, k, len, sizhalf,first_point,last_point;
float t1,t2;
int ia,ib,ic;
len=lk|1;
sizhalf=mix2_size/2;
first_point=(lapt-sizhalf+baseband_size)&baseband_mask;
ia=(first_point+1-len+baseband_size)&baseband_mask;
ib=(first_point-len/2+baseband_size)&baseband_mask;
ic=first_point+1;
t1=0;
t2=0;
k=ia;
for(i=0; i<len; i++)
  {
  t1+=xin[2*k];
  t2+=xin[2*k+1];
  k=(k+1)&baseband_mask;
  }
last_point=(sizhalf+first_point)&baseband_mask;
while(ic!=last_point)
  {
  xout[2*ib  ]=t1/len;
  xout[2*ib+1]=t2/len;
  t1+=xin[2*ic  ];
  t2+=xin[2*ic+1];
  ic=(ic+1)&baseband_mask;
  t1-=xin[2*ia  ];
  t2-=xin[2*ia+1];
  ib=(ib+1)&baseband_mask;
  ia=(ia+1)&baseband_mask;
  }
xout[2*ib  ]=t1/len;
xout[2*ib+1]=t2/len;
return (ib+1)&baseband_mask;
}


